@@ -36,6 +36,7 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 #include <X11/Xatom.h>
+#include <X11/extensions/dpms.h>
 
 #undef GNOME_DISABLE_DEPRECATED
 #include "gnome-rr.h"
@@ -100,6 +101,8 @@ struct GnomeRROutput
     guint8 *		edid_data;
     gsize		edid_size;
     char *              connector_type;
+    gint		backlight_min;
+    gint		backlight_max;
 };
 
 struct GnomeRROutputWrap
@@ -530,6 +533,11 @@ fill_out_screen_info (Display *xdisplay,
     }
 #endif
 
+    /* can the screen do DPMS? */
+    gdk_error_trap_push ();
+    priv->dpms_capable = DPMSCapable (priv->xdisplay);
+    gdk_error_trap_pop_ignored ();
+
     return TRUE;
 #else
     return FALSE;
@@ -1163,6 +1171,125 @@ gnome_rr_screen_refresh (GnomeRRScreen *screen,
     return refreshed;
 }
 
+/**
+ * gnome_rr_screen_get_dpms_mode:
+ **/
+gboolean
+gnome_rr_screen_get_dpms_mode (GnomeRRScreen *screen,
+                               GnomeRRDpmsMode *mode,
+                               GError **error)
+{
+    BOOL enabled = FALSE;
+    CARD16 state;
+    gboolean ret = FALSE;
+
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+    g_return_val_if_fail (mode != NULL, FALSE);
+
+    if (!screen->priv->dpms_capable) {
+        g_set_error_literal (error,
+                             GNOME_RR_ERROR,
+                             GNOME_RR_ERROR_NO_DPMS_EXTENSION,
+                             "Display is not DPMS capable");
+        goto out;
+    }
+
+    if (!DPMSInfo (screen->priv->xdisplay,
+                   &state,
+                   &enabled)) {
+        g_set_error_literal (error,
+                             GNOME_RR_ERROR,
+                             GNOME_RR_ERROR_UNKNOWN,
+                             "Unable to get DPMS state");
+        goto out;
+    }
+
+    /* DPMS not enabled is a valid mode */
+    if (!enabled) {
+        *mode = GNOME_RR_DPMS_DISABLED;
+        ret = TRUE;
+        goto out;
+    }
+
+    switch (state) {
+    case DPMSModeOn:
+        *mode = GNOME_RR_DPMS_ON;
+        break;
+    case DPMSModeStandby:
+        *mode = GNOME_RR_DPMS_STANDBY;
+        break;
+    case DPMSModeSuspend:
+        *mode = GNOME_RR_DPMS_SUSPEND;
+        break;
+    case DPMSModeOff:
+        *mode = GNOME_RR_DPMS_OFF;
+        break;
+    default:
+        g_assert_not_reached ();
+        break;
+    }
+    ret = TRUE;
+out:
+    return ret;
+}
+
+/**
+ * gnome_rr_screen_set_dpms_mode:
+ **/
+gboolean
+gnome_rr_screen_set_dpms_mode (GnomeRRScreen *screen,
+                               GnomeRRDpmsMode mode,
+                               GError **error)
+{
+    CARD16 state = 0;
+    gboolean ret;
+    GnomeRRDpmsMode current_mode;
+
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+    /* set, if the new mode is different */
+    ret = gnome_rr_screen_get_dpms_mode (screen, &current_mode, error);
+    if (!ret)
+        goto out;
+    if (current_mode == mode)
+        goto out;
+
+    switch (mode) {
+    case GNOME_RR_DPMS_ON:
+        state = DPMSModeOn;
+        break;
+    case GNOME_RR_DPMS_STANDBY:
+        state = DPMSModeStandby;
+        break;
+    case GNOME_RR_DPMS_SUSPEND:
+        state = DPMSModeSuspend;
+        break;
+    case GNOME_RR_DPMS_OFF:
+        state = DPMSModeOff;
+        break;
+    default:
+        g_assert_not_reached ();
+        break;
+    }
+
+    gdk_error_trap_push ();
+    ret = DPMSForceLevel (screen->priv->xdisplay, state);
+    gdk_flush ();
+    if (gdk_error_trap_pop ())
+        ret = FALSE;
+
+    if (!ret) {
+        ret = FALSE;
+        g_set_error_literal (error,
+                             GNOME_RR_ERROR,
+                             GNOME_RR_ERROR_UNKNOWN,
+                             "Could not change DPMS mode");
+        goto out;
+    }
+out:
+    return ret;
+}
+
 /**
  * gnome_rr_screen_list_modes:
  *
@@ -1399,6 +1526,46 @@ out:
 #endif
 }
 
+static void
+update_brightness_limits (GnomeRROutput *output)
+{
+#ifdef HAVE_RANDR
+    gint rc;
+    Atom atom;
+    XRRPropertyInfo *info;
+
+    gdk_error_trap_push ();
+    atom = XInternAtom (DISPLAY (output), "BACKLIGHT", FALSE);
+    info = XRRQueryOutputProperty (DISPLAY (output), output->id, atom);
+    rc = gdk_error_trap_pop ();
+    if (rc != Success)
+    {
+        if (rc != BadName)
+          g_warning ("could not get output property for %s, rc: %i",
+		     output->name, rc);
+        goto out;
+    }
+    if (info == NULL)
+    {
+        g_warning ("could not get output property for %s",
+		   output->name);
+        goto out;
+    }
+    if (!info->range || info->num_values != 2)
+    {
+        g_debug ("backlight %s was not range", output->name);
+        goto out;
+    }
+    output->backlight_min = info->values[0];
+    output->backlight_max = info->values[1];
+out:
+    if (info != NULL)
+    {
+        XFree (info);
+    }
+#endif
+}
+
 #ifdef HAVE_RANDR
 static gboolean
 output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **error)
@@ -1470,7 +1637,11 @@ output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **erro
     
     /* Edid data */
     output->edid_data = read_edid_data (output, &output->edid_size);
-    
+
+    /* brightness data */
+    if (output->connected)
+        update_brightness_limits (output);
+
     XRRFreeOutputInfo (info);
 
     return TRUE;
@@ -1495,6 +1666,8 @@ output_copy (const GnomeRROutput *from)
     output->connected = from->connected;
     output->n_preferred = from->n_preferred;
     output->connector_type = g_strdup (from->connector_type);
+    output->backlight_min = -1;
+    output->backlight_max = -1;
 
     array = g_ptr_array_new ();
     for (p_crtc = from->possible_crtcs; *p_crtc != NULL; p_crtc++)
@@ -1552,6 +1725,153 @@ gnome_rr_output_get_edid_data (GnomeRROutput *output, gsize *size)
     return output->edid_data;
 }
 
+/**
+ * gnome_rr_output_get_backlight_min:
+ *
+ * Returns: The mimimum backlight value, or -1 if not supported
+ */
+gint
+gnome_rr_output_get_backlight_min (GnomeRROutput *output)
+{
+    g_return_val_if_fail (output != NULL, -1);
+    return output->backlight_min;
+}
+
+/**
+ * gnome_rr_output_get_backlight_max:
+ *
+ * Returns: The maximum backlight value, or -1 if not supported
+ */
+gint
+gnome_rr_output_get_backlight_max (GnomeRROutput *output)
+{
+    g_return_val_if_fail (output != NULL, -1);
+    return output->backlight_max;
+}
+
+/**
+ * gnome_rr_output_get_backlight:
+ *
+ * Returns: The currently set backlight brightness
+ */
+gint
+gnome_rr_output_get_backlight (GnomeRROutput *output, GError **error)
+{
+    guint now = -1;
+#ifdef HAVE_RANDR
+    unsigned long nitems;
+    unsigned long bytes_after;
+    guint *prop;
+    Atom atom;
+    Atom actual_type;
+    int actual_format;
+    gint retval;
+
+    g_return_val_if_fail (output != NULL, -1);
+
+    gdk_error_trap_push ();
+    atom = XInternAtom (DISPLAY (output), "BACKLIGHT", FALSE);
+    retval = XRRGetOutputProperty (DISPLAY (output), output->id, atom,
+				   0, 4, False, False, None,
+				   &actual_type, &actual_format,
+				   &nitems, &bytes_after, ((unsigned char **)&prop));
+    gdk_flush ();
+    if (gdk_error_trap_pop ())
+    {
+        g_set_error_literal (error,
+			     GNOME_RR_ERROR,
+			     GNOME_RR_ERROR_UNKNOWN,
+			     "unhandled X error while getting the range of backlight values");
+        goto out;
+    }
+
+    if (retval != Success) {
+        g_set_error_literal (error,
+			     GNOME_RR_ERROR,
+			     GNOME_RR_ERROR_RANDR_ERROR,
+			     "could not get the range of backlight values");
+        goto out;
+    }
+    if (actual_type == XA_INTEGER &&
+        nitems == 1 &&
+        actual_format == 32)
+    {
+        memcpy (&now, prop, sizeof (guint));
+    }
+    else
+    {
+	g_set_error (error,
+		     GNOME_RR_ERROR,
+		     GNOME_RR_ERROR_RANDR_ERROR,
+		     "failed to get correct property type, got %lu,%i",
+		     nitems, actual_format);
+    }
+out:
+    XFree (prop);
+#else
+    g_set_error_literal (error,
+			 GNOME_RR_ERROR,
+			 GNOME_RR_ERROR_NO_RANDR_EXTENSION,
+			 "not compiled with RANDR support");
+#endif /* HAVE_RANDR */
+    return now;
+}
+
+/**
+ * gnome_rr_output_set_backlight:
+ * @value: the absolute value which is min >= this <= max
+ *
+ * Returns: %TRUE for success
+ */
+gboolean
+gnome_rr_output_set_backlight (GnomeRROutput *output, gint value, GError **error)
+{
+    gboolean ret = FALSE;
+#ifdef HAVE_RANDR
+    Atom atom;
+
+    g_return_val_if_fail (output != NULL, FALSE);
+
+    /* check this is sane */
+    if (value < output->backlight_min ||
+        value > output->backlight_max)
+    {
+	g_set_error (error,
+		     GNOME_RR_ERROR,
+		     GNOME_RR_ERROR_BOUNDS_ERROR,
+		     "out of brightness range: %i, has to be %i -> %i",
+		     value,
+		     output->backlight_max, output->backlight_min);
+	goto out;
+    }
+
+    /* don't abort on error */
+    gdk_error_trap_push ();
+    atom = XInternAtom (DISPLAY (output), "BACKLIGHT", FALSE);
+    XRRChangeOutputProperty (DISPLAY (output), output->id, atom,
+			     XA_INTEGER, 32, PropModeReplace,
+			     (unsigned char *) &value, 1);
+    if (gdk_error_trap_pop ())
+    {
+        g_set_error_literal (error,
+			     GNOME_RR_ERROR,
+			     GNOME_RR_ERROR_UNKNOWN,
+			     "unhandled X error while setting the backlight values");
+        goto out;
+    }
+
+    /* we assume this succeeded as there's no return value */
+    ret = TRUE;
+#else
+    g_set_error_literal (error,
+			 GNOME_RR_ERROR,
+			 GNOME_RR_ERROR_NO_RANDR_EXTENSION,
+			 "not compiled with RANDR support");
+#endif /* HAVE_RANDR */
+out:
+    return ret;
+}
+
 /**
  * gnome_rr_screen_get_output_by_id:
  *
